####                                       ####<< methods >>####
####                                       ####< gff handling >####
#### extract attribute fields from refFlat_hg18/19.gtf
getAttributeFieldRefFlat <- function(x, field, attrsep = "; ") {
  s = strsplit(x, split = attrsep, fixed = TRUE)
  res= sapply(s, function(atts) {
    a = strsplit(atts, split = " ", fixed = TRUE)
    m = match(field, sapply(a, "[", 1))
    if (!is.na(m)) {
      rv = a[[m]][2]
    }
    else {
      rv = as.character(NA)
    }
    return(rv)
  })
  res=unlist(lapply(strsplit(res,split="\""),function(el){el[2]}))
}
#### gff parser
gffRead <- function(gffFile, nrows = -1,header=FALSE) {
  cat("Reading ", gffFile, ": ", sep="")
  gff = read.table(gffFile, sep="\t", as.is=TRUE, quote="",header=header, comment.char="#", nrows = nrows,
    colClasses=c("character", "character", "character", "integer",
      "integer","character", "character", "character", "character")
    )
  colnames(gff) = c("seqname", "source", "feature", "start", "end", "score", "strand", "frame", "attributes")
  cat("found", nrow(gff), "rows with classes:",
      paste(sapply(gff, class), collapse=", "), "\n")
  stopifnot(!any(is.na(gff$start)), !any(is.na(gff$end)))
  return(gff)
}
####                                       ####< addGeneSymbolToAnnot >####
readAndFixIdToSymbolFile<-function(idToSymbolFile){
  idToSymbol=read.table(idToSymbolFile,header=FALSE,sep="\t",colClasses=c("character","character"))
  idToSymbol[,2]=gsub(",$","",idToSymbol[,2])
  return(idToSymbol)
}
#### adds geneSymbol to annotation based on mapping provided in idToSymbolFile
#### annot: must contain transcript_id column
#### idToSymbolFile: must contain two columns. First is transcript_id, second is geneSymbol
addGeneSymbolToAnnot<-function(annot,idToSymbolFile,logFile){
  idToSymbol=readAndFixIdToSymbolFile(idToSymbolFile)
  colnames(idToSymbol)=c("ID","geneSymbol")
  idToSymbol=idToSymbol[!duplicated(idToSymbol),]
  
  if(sum(duplicated(idToSymbol[,"ID"]))){
    stop("some transcripts are assigned to different genes in idToSymbolFile")
  }
  
  m=merge(annot,idToSymbol,by.x="transcript_id",by.y="ID",all.x=TRUE,all.y=FALSE)

  cat(paste("removing",sum(is.na(m[,"geneSymbol"])),"cases without geneSymbol\n"))
  if (any(is.na(m[,"geneSymbol"]))){
    cat("addGeneSymbolToAnnot removed the following cases without geneSymbol:\n",file=logFile,append=TRUE)
    cat(m[is.na(m[,"geneSymbol"]),"transcript_id"],file=logFile,append=TRUE)
  }
  
  m=m[!is.na(m[,"geneSymbol"]),]
  return(m)
}
####                                       ####<<   >>####
#### removes transcripts that are annotated on multiple chromosomes
removeTranscriptsWithAmbigChr<-function(annot,logFile){
  seqName=unique(subset(annot,select=c("transcript_id","seqname")))
  dupIDs=seqName[duplicated(seqName[,"transcript_id"]),"transcript_id"] ## ids with multiple chromosomes
  cat(c("removing",length(dupIDs),"transcripts\n"))
  cat("removeTranscriptsWithAmbigChr removed the following transcripts:\n",file=logFile,append=TRUE)
  cat(dupIDs,file=logFile,append=TRUE)
  cat("\n ======================== \n",file=logFile,append=TRUE)
  annot=subset(annot,!(transcript_id %in% dupIDs))
  return(annot)
}
#### merges all transcripts in annot that belong to qGeneSymbol and returns a union transcript
mergeTranscriptsOfGene<-function(annot,qGeneSymbol,logFile){
  if (is.na(qGeneSymbol)){
    return(data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL,mergeComment=NULL))
  }
  gff=subset(annot,geneSymbol == qGeneSymbol)
  if (nrow(gff)==0){
    return(data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL,mergeComment=NULL))
  }

  res_seqname=uniqueSingle(gff$seqname,qGeneSymbol,logFile)
  res_source=uniqueSingle(gff$source,qGeneSymbol,logFile)
  res_feature=uniqueSingle(gff$feature,qGeneSymbol,logFile)
  res_score=uniqueSingle(gff$score,qGeneSymbol,logFile)
  res_strand=uniqueSingle(gff$strand,qGeneSymbol,logFile)
  res_frame="."##uniqueSingle(gff$frame,qGeneSymbol,logFile)
  res_attributes=paste(unique(gff$transcript_id),collapse=";")
  res_geneSymbol=qGeneSymbol
  res_mergeComment=""
  
                                        #+ test for disjunctive transcripts
  gff[,"start_end"]=paste(gff[,"start"],gff[,"end"],sep="_")
  s=split(gff,list(gff[,"transcript_id"]))
  l=lapply(s,function(el){return(el[,"start_end"])})
  inter=.mulitpleIntersect(l) 
  res=subset(s[[1]],start_end %in% inter,select=c(start,end)) ## only exons which have the same annotated start and end in all transcripts
  if (nrow(res)==0){ ## transcripts have no exon in common e.g. UNKL
    ##warning(qGeneSymbol," has disjunct transcripts")
    cat(paste("mergeTranscriptsOfGene:",qGeneSymbol," has disjunct transcripts\n"),file=logFile,append=TRUE)
    res_mergeComment="disjunctTranscripts"
  }
                                        #+ --------------------------------

  a=cbind(gff[,c("start")],1)
  b=cbind(gff[,c("end")],-1)
  colnames(a)=colnames(b)=c("pos","trans")
  gff.s=rbind(a,b) # create +1 for start and -1 for end
  gff.s=gff.s[order(gff.s[,"pos"]),] # sort
  cs=cumsum(gff.s[,"trans"]) # cumsum is a count of the transitions
  gff.s=cbind(gff.s,cs)
  
  res=cbind(gff.s[gff.s[,"trans"]==1 & gff.s[,"cs"]==1,"pos"],gff.s[gff.s[,"trans"]==-1 & gff.s[,"cs"]==0,"pos"])
  rslt=data.frame(res_seqname,res_source,res_feature,res,res_score,res_strand,res_frame,res_attributes,res_geneSymbol,res_mergeComment)
  names(rslt)=c("seqname","source","feature","start","end","score","strand","frame","attributes","geneSymbol","mergeComment")
  return(rslt)
}
#### merges all transcripts in annot that belong to qGeneSymbol and returns a intersection transcript
mergeTranscriptsOfGeneIntersection<-function(annot,qGeneSymbol){
  if (is.na(qGeneSymbol)){
    return(data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL,mergeComment=NULL))
  }
  gff=subset(annot,geneSymbol == qGeneSymbol)
  if (nrow(gff)==0){
    return(data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL,mergeComment=NULL))
  }

  res_seqname=uniqueSingle(gff$seqname,qGeneSymbol,logFile)
  res_source=uniqueSingle(gff$source,qGeneSymbol,logFile)
  res_feature=uniqueSingle(gff$feature,qGeneSymbol,logFile)
  res_score=uniqueSingle(gff$score,qGeneSymbol,logFile)
  res_strand=uniqueSingle(gff$strand,qGeneSymbol,logFile)
  res_frame="."##uniqueSingle(gff$frame,qGeneSymbol,logFile)
  res_attributes=paste(unique(gff$transcript_id),collapse=";")
  res_geneSymbol=qGeneSymbol
  res_mergeComment=""
  
  gff[,"start_end"]=paste(gff[,"start"],gff[,"end"],sep="_")
  s=split(gff,list(gff[,"transcript_id"]))
  l=lapply(s,function(el){return(el[,"start_end"])})
  inter=.mulitpleIntersect(l) 
  res=subset(s[[1]],start_end %in% inter,select=c(start,end)) ## only exons which have the same annotated start and end in all transcripts
  if (nrow(res)==0){ ## transcripts have no exon in common e.g. UNKL
    ##warning(qGeneSymbol," has disjunct transcripts")
    cat(paste("mergeTranscriptsOfGeneIntersection:",qGeneSymbol,"has disjunct transcripts\n"),file=logFile,append=TRUE)
    return(data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL,mergeComment=NULL))
  }

  rslt=data.frame(res_seqname,res_source,res_feature,res,res_score,res_strand,res_frame,res_attributes,res_geneSymbol,res_mergeComment)
  rslt=rslt[order(rslt[,"start"]),]
  names(rslt)=c("seqname","source","feature","start","end","score","strand","frame","attributes","geneSymbol","mergeComment")
  return(rslt)
}
.mulitpleIntersect<-function(listOfSets){
  if (length(listOfSets)==1){
    return(listOfSets[[1]])
  }
  tmp=intersect(listOfSets[[1]],listOfSets[[2]])
  if (length(listOfSets)>2){
    for (i in 3:length(listOfSets)){
      tmp=intersect(tmp,listOfSets[[i]])
    }
  }
  return(tmp)
}
####                                       ####< reformat >####
#### reformat table into gff format
reformatToGff<-function(tab){
  res=subset(tab,select=c(seqname,source,feature,start,end,score,strand,frame))
  mergedTranscriptIDs=tab$attributes
  mergedTranscriptIDs=sub(";","_",mergedTranscriptIDs)
  res$attributes=paste("geneSymbol \"",tab$geneSymbol,"\"; mergedTranscriptIDs \"",mergedTranscriptIDs,"\";",sep="")
  res$end=as.integer(res$end)
  res$start=as.integer(res$start)
  return(res)
}
#### returns single value. Either unique value or collapsed value (if val contains different entries).
uniqueSingle<-function(val,geneSymbol,logFile){
  res=unique(val)
  if (length(res)>1){
    ##warning(c(geneSymbol,": ",res," ambiguous"))
    cat(paste(geneSymbol,": ",res," ambiguous\n"),file=logFile,append=TRUE)
  }
  return(paste(res,collapse=""))
}
####                                       ####< main >####
#### checks if annot semms ok before doing the merge
checkAnnot<-function(annot){
  
  t1=all(c("seqname","source","feature","start","end","score","strand","frame","attributes","geneSymbol") %in% colnames(annot))
  t2=all(c("transcript_id","geneSymbol") %in% colnames(annot))

  t3=length(unique(annot$feature))==1
  t4=all(annot$start<annot$end)

  seqName=unique(subset(annot,select=c("transcript_id","seqname")))
  dupIDs=seqName[duplicated(seqName[,"transcript_id"]),"transcript_id"] 
  t5=length(dupIDs)==0

  return(t1 && t2 && t3 && t4 && t5)
}
#### for every gene in annot (as identified by a symbol), transcripts are merged together
#### result is saved in outFolder/outFileBaseName[Intersection|Union].gtf
mergeAndSave<-function(annot,outFolder,outFileBaseName,method="Intersection",logFile){
  stopifnot(method %in% c("Union","Intersection"))
  if(!checkAnnot(annot)) stop("please check annotation")

  outFileName=paste(outFolder,outFileBaseName,method,".gtf",sep="")
  print(paste("generating",outFileName,sep=""))

  geneSymbols=unique(annot$geneSymbol)
  res=data.frame(seqname=NULL,source=NULL,feature=NULL,start=NULL,end=NULL,score=NULL,strand=NULL,frame=NULL,attributes=NULL,geneSymbol=NULL)
  pb=txtProgressBar(min=0,max=length(geneSymbols),style=3)
  for (i in 1:length(geneSymbols)){
    setTxtProgressBar(pb,i)
    geneSymbol=geneSymbols[i]
    if (method=="Intersection") res=rbind(res,mergeTranscriptsOfGeneIntersection(annot,geneSymbol,logFile))
    if (method=="Union") res=rbind(res,mergeTranscriptsOfGene(annot,geneSymbol,logFile))
  }
  close(pb)

  cat(paste("writing",outFileName,"\n"))
  res=reformatToGff(res)
  write.table(res,outFileName,col.names=FALSE,row.names=FALSE,sep="\t",quote=FALSE)

  return(res)
}
####                                       ####<< synopsis >>####
synopsis<-function(){
                                        #+ please provide gtf file and mapping from transcript id to gene symbol
  annotFile="hf19_annotation.gtf" ## reference annotation file
  idToSymbolFile="hf19_idToSymbol.txt" ## mapping file
  logFile="ensGeneTable_exon.log" ## name of output log file
  outFileBaseName="ensGeneTable_exon" ## basename of output file

                                          #+ prepare input annotation
  annot=gffRead(annotFile)
  annot=subset(annot,feature=="exon")
  ## you can also choose ‘‘annot=subset(annot,feature==’’exon’’)’’
  ## if you want UTRs represented in your union transcripts
  
  annot$transcript_id=getAttributeFieldRefFlat(annot$attributes,"gene_id")
  ## In CCDS (as downloaded from UCSC-Browser), transcript_id is equal to gene_id
  ## In RefSeq (as downloaded from UCSC-Browser), transcript_id differs from gene_id.
  ## For mapping purposes of this script please use the above syntax.
  
  annot=addGeneSymbolToAnnot(annot,idToSymbolFile,logFile)

  annot=removeTranscriptsWithAmbigChr(annot,logFile)
  ## This should be improved sometime because it removes transcripts
  ## that are associated to multiple chromosomes (see log file)
  
  annot=annot[annot$start<annot$end,] ## filters some weird entries

  checkAnnot(annot) ## must return TRUE

                                        #+ merge and save
  ## the following produces the files
  ## ‘‘[outFolder]/[outFileBaseName]Union.gtf’’ and ‘‘[logFile]’’:
  res=mergeAndSave(annot,
    outFolder="output/",
    outFileBaseName=outFileBaseName,
    method="Union",
    logFile=logFile)
}
